import collections
import csv
import datetime
import io

from django.utils import timezone


def create_csv_from_data(data, stream=None):
    """Creates a CSV stream using the given dict, in where the keys are
    the columns and each value is a list of results.
    """
    stream = io.StringIO() if not stream else stream
    header = list(data.keys())
    writer = csv.DictWriter(stream, fieldnames=data.keys())
    writer.writeheader()
    for index in range(len(data[header[0]])):
        row = {key: data[key][index] for key in header}
        writer.writerow(row)
    return stream


def age_range_filter(field, min_age=None, max_age=None):
    """Returns the filter for the age range."""
    current = timezone.now().date()
    _filter = {}
    if min_age:
        _filter[f"{field}__year__lte"] = datetime.date(
            current.year - min_age, current.month, current.day
        ).year
    if max_age:
        _filter[f"{field}__year__gte"] = datetime.date(
            current.year - max_age, current.month, current.day
        ).year
    return _filter


StatisticsValue = collections.namedtuple(
    "StatisticsValue", ["value", "percentage", "help_text"]
)
StatisticsValue.__new__.__defaults__ = (False, None)



from django.conf import settings


# Needed to build and publish
# ------------------------------------------------------------------------------
SECRET_KEY = "backoffice"

# Specific project configuration
# ------------------------------------------------------------------------------
TITLE = getattr(settings, "BACKOFFICE_TITLE", "backoffice")
URL_NAMESPACE = getattr(settings, "BACKOFFICE_URL_NAMESPACE", "backoffice")
BOOLEAN_TRUE_ICON_CLASSES = getattr(
    settings,
    "BACKOFFICE_BOOLEAN_TRUE_ICON_CLASSES",
    "has-text-success fas fa-check-circle",
)
BOOLEAN_FALSE_ICON_CLASSES = getattr(
    settings,
    "BACKOFFICE_BOOLEAN_FALSE_ICON_CLASSES",
    "has-text-danger fas fa-times-circle",
)

STATUS_FIELDS = getattr(settings, "BACKOFFICE_STATUS_FIELDS", ("status",))
STATUS_TAG_CLASSES = getattr(settings, "BACKOFFICE_STATUS_TAG_CLASSES", {})
DETAILS_URLS = getattr(
    settings,
    "BACKOFFICE_DETAILS_URLS",
    [{"names": ("pk", "id")}, {"names": ("user", "owner")},],
)
SIDEBAR_CONFIG = getattr(settings, "BACKOFFICE_SIDEBAR_CONFIG", {})



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class BackofficeAppConfig(AppConfig):

    name = "backoffice_extensions"
    verbose_name = _("Backoffice")



import collections
from functools import reduce
from typing import AnyStr, Dict, List, Optional, Type

from django.db import models
from django.db.models import Q
from django.http import HttpResponse
from django.urls import NoReverseMatch, reverse

from backoffice_extensions.helpers import create_csv_from_data
from backoffice_extensions.settings import TITLE, URL_NAMESPACE


class BackOfficeViewMixin:
    template_name: AnyStr = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.template_name:
            raise NotImplementedError("You should specify the template_name attribute.")

    def get_extra_context(self) -> Dict:
        """Adds default context to the backoffice views. Overwrite to add more 
        context to the view.
        """
        try:
            index_url = reverse(f"{URL_NAMESPACE}:index")
        except NoReverseMatch:
            index_url = ""
        try:
            sign_out = reverse(f"{URL_NAMESPACE}:sign-out")
        except NoReverseMatch:
            sign_out = ""
        try:
            sign_in = reverse(f"{URL_NAMESPACE}:sign-in")
        except NoReverseMatch:
            sign_in = ""
        return {
            "backoffice_title": TITLE,
            "index_url": index_url,
            "sign_out": sign_out,
            "sign_in": sign_in,
        }


class SearchListMixin:
    """Mixin to add search functionality to default ListView
    Django view."""

    search_param: AnyStr = "search"
    search_fields: List = []

    def _search_filter(self, queryset) -> models.QuerySet:
        """Applies search filtering to queryset."""
        search_query = self.request.GET.get(self.search_param)
        if search_query:
            queryset_filter = [
                Q(**{f"{search_field}__icontains": search_query})
                for search_field in self.search_fields
            ]
            if queryset_filter:
                queryset = queryset.filter(reduce(lambda x, y: x | y, queryset_filter))
        return queryset

    def get_queryset(self) -> models.QuerySet:
        """Checks the 'search' variable to allow generic search."""
        queryset = super().get_queryset()
        queryset = self._search_filter(queryset)
        return queryset

    def get_extra_context(self) -> Dict:
        """Adds search data."""
        context: Dict = super().get_extra_context()
        context[self.search_param] = self.request.GET.get(self.search_param, "")
        parameters = self.request.GET.copy()
        parameters = parameters.pop(self.page_kwarg, True) and parameters.urlencode()
        context["parameters"] = parameters
        return context


class ExportMixin:
    """Mixin to allow export CSV data."""

    filename: AnyStr = "data.csv"
    queryset: models.QuerySet = None
    filterset_class = None
    fields: List = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.queryset:
            raise NotImplementedError("You should specify the queryset attribute.")

    def get_csv_response(self, data):
        response = HttpResponse(content_type="text/csv")
        response[
            "Content-Disposition"
        ] = f'attachment; filename="{self.get_filename()}"'
        create_csv_from_data(data, stream=response)
        return response

    def get_filename(self) -> AnyStr:
        return self.filename

    def get_queryset(self) -> models.QuerySet:
        return self.queryset

    def get(self, request, *args, **kwargs):
        items: models.QuerySet = self.get_queryset()
        if self.filterset_class:
            _filter = self.filterset_class(request.GET, queryset=items)
            items = _filter.qs
        fields: List = [
            field[0] if isinstance(field, tuple) else field for field in self.fields
        ]
        if "id" not in fields:
            fields = ["id"] + fields
        data: collections.OrderedDict = collections.OrderedDict()
        for field in fields:
            data[field] = []
        for item in items.iterator():
            for field in fields:
                value = getattr(item, field)
                if hasattr(value, "__call__"):
                    value = value()
                data[field].append(value)
        return self.get_csv_response(data=data)



from django import template
from django.apps import apps
from django.contrib.gis.geos import Point
from django.contrib.humanize.templatetags.humanize import intcomma
from django.db.models import FieldDoesNotExist, Manager, QuerySet
from django.db.models.fields.files import ImageFieldFile
from django.template import defaultfilters
from django.urls import NoReverseMatch, reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from backoffice_extensions.helpers import StatisticsValue
from backoffice_extensions.settings import (
    BOOLEAN_FALSE_ICON_CLASSES,
    BOOLEAN_TRUE_ICON_CLASSES,
    DETAILS_URLS,
    STATUS_FIELDS,
    STATUS_TAG_CLASSES,
    URL_NAMESPACE,
    SIDEBAR_CONFIG,
)

register = template.Library()


@register.inclusion_tag("backoffice/partials/menu.html", takes_context=True)
def sidebar_menu(context):
    """Creates the sidebar data."""
    user = context.get("user")
    sidebar = []
    for section, entries in SIDEBAR_CONFIG.items():
        entries_data = []
        for entry, data in entries.items():
            if data.get("permission") is None or (
                user and user.has_perm(data.get("permission"))
            ):
                entries_data.append(
                    (
                        reverse(f"{URL_NAMESPACE}:{entry.lower()}-list"),
                        data.get("label"),
                    )
                )
        sidebar.append((section, entries_data))
    return {"sidebar": sidebar}


@register.filter
def boolean_icon(value):
    """Gets an icon for given boolean value."""
    result = f'<i class="{BOOLEAN_FALSE_ICON_CLASSES}"></i>'
    if value:
        result = f'<i class="{BOOLEAN_TRUE_ICON_CLASSES}"></i>'
    return mark_safe(result)


@register.filter
def status_tag(value):
    """Gets an icon for given boolean value."""
    result = (
        f'<span class="tag {STATUS_TAG_CLASSES.get(value.status, "")}">'
        f"{value.get_status_display()}</i>"
    )
    return mark_safe(result)


@register.filter(name="getattr")
def getattr_filter(obj, name):
    """Calls to getattr over the given obj with the given name."""

    def _parse_value(value):
        """Parse the given value."""
        if value is None:
            value = "-"
        if isinstance(value, bool):
            return boolean_icon(value)
        if isinstance(value, Point):
            return f"{value.y},{value.x}"
        if isinstance(value, ImageFieldFile):
            if value:
                return mark_safe(f'<img src="{value.url}" />')
            return _("No picture")
        if isinstance(value, Manager):
            value = ", ".join([str(item) for item in value.all()]) or "-"
        return value

    if isinstance(name, tuple) and len(name) > 0:
        name = name[0]
    result = getattr(obj, name)
    if result is None:
        result = "-"
    for detail_url_data in DETAILS_URLS:
        names = detail_url_data.get("names", tuple())
        lookup_field = detail_url_data.get("lookup_field") or "pk"
        lookup_field_value = (
            getattr(result, lookup_field) if hasattr(result, lookup_field) else result
        )
        if name in names:
            try:
                details_url = reverse(
                    f"{URL_NAMESPACE}:{obj._meta.model_name}-detail",
                    kwargs={lookup_field: lookup_field_value},
                )
                result = mark_safe(f'<a href="{details_url}">{str(result)}</a>')
            except (NoReverseMatch, AttributeError):
                result = result
    if name in STATUS_FIELDS:
        result = status_tag(obj)
    if hasattr(result, "__call__") and not isinstance(result, Manager):
        result = result()
    return _parse_value(result)


@register.filter
def verbose_name(model_or_queryset, field):
    """Gets the verbose name of the given model field."""
    model = model_or_queryset
    if isinstance(field, tuple) and len(field) > 1:
        return field[1]
    if isinstance(model_or_queryset, QuerySet):
        model = model_or_queryset.model
    try:
        return model._meta.get_field(field).verbose_name
    except FieldDoesNotExist:
        if hasattr(model, field) and hasattr(getattr(model, field), "verbose_name"):
            return getattr(getattr(model, field), "verbose_name")
    return field


@register.filter
def statistics_value(value):
    if not isinstance(value, StatisticsValue):
        return intcomma(value)
    result = f"{intcomma(value.value)}"
    if value.percentage:
        result = f"{defaultfilters.floatformat(value.value, 2)} %"
    return result



"""Django app made to create backoffices to help the administration of a site, like 
the Django Admin site, but for final users.
"""
from django.utils.module_loading import autodiscover_modules

__version__ = "1.0.1"


default_app_config = "backoffice_extensions.apps.BackofficeAppConfig"



from typing import AnyStr, Dict, List, Optional, Type

from django import forms
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import LoginRequiredMixin
from django.db import models
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import ListView

from backoffice_extensions.mixins import BackOfficeViewMixin
from backoffice_extensions.settings import URL_NAMESPACE

User = get_user_model()


class BackOfficeFormView(LoginRequiredMixin, BackOfficeViewMixin, View):
    """Base view for forms."""

    form_class: Type[forms.ModelForm] = forms.ModelForm

    def get_model_class(self) -> Type[models.Model]:
        """Extracts the model class form the ModelFrom."""
        return self.form_class._meta.model

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.form_class or not issubclass(self.form_class, forms.ModelForm):
            raise NotImplementedError(
                "You should specify the form_class attribute, and it have to be a "
                "subclass of django.forms.ModelForm"
            )


class BackOfficeCreateView(BackOfficeFormView):
    """Base view for creations."""

    success_message: AnyStr = _("{instance} created")

    def get(self, request):
        form = self.form_class()
        context = {"form": form}
        context.update(self.get_extra_context())
        return render(request, self.template_name, context=context)

    def post(self, request):
        model_class = self.get_model_class()
        form = self.form_class(request.POST, request.FILES)
        context = {"form": form}
        context.update(self.get_extra_context())
        if form.is_valid():
            instance = form.save()
            messages.success(
                request, self.success_message.format(instance=str(instance))
            )
            return redirect(
                f"{URL_NAMESPACE}:{model_class._meta.model_name}-detail", pk=instance.pk
            )
        return render(request, self.template_name, context=context)


class BackOfficeEditView(BackOfficeFormView):
    """Base view for editions."""

    success_message = _("{instance} updated")

    def get(self, request, pk):
        model_class = self.get_model_class()
        instance = get_object_or_404(model_class, pk=pk)
        form = self.form_class(instance=instance)
        context = {"form": form, "instance": instance}
        context.update(self.get_extra_context())
        return render(request, self.template_name, context=context)

    def post(self, request, pk):
        model_class = self.get_model_class()
        instance = get_object_or_404(model_class, pk=pk)
        form = self.form_class(request.POST, request.FILES, instance=instance)
        context = {"form": form, "instance": instance}
        context.update(self.get_extra_context())
        if form.is_valid():
            instance = form.save()
            messages.success(
                request, self.success_message.format(instance=str(instance))
            )
            return redirect(
                f"{URL_NAMESPACE}:{model_class._meta.model_name}-detail", pk=instance.pk
            )
        return render(request, self.template_name, context=context)


class BackOfficeListView(LoginRequiredMixin, BackOfficeViewMixin, ListView):
    """Base view for lists."""

    list_display: List = []
    filterset_class = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filter = None

    def get_queryset(self):
        """Uses the FilterSet class to filter the query."""
        queryset = super().get_queryset()
        if self.filterset_class:
            self.filter = self.filterset_class(self.request.GET, queryset=queryset)
            queryset = self.filter.qs
        return queryset

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context.update({"list_display": self.list_display, "filter": self.filter})
        context.update(self.get_extra_context())
        return context


class BackOfficeDetailView(LoginRequiredMixin, BackOfficeViewMixin, View):
    """Base detail view."""

    queryset: Optional[models.QuerySet] = None
    model_class: Type[models.Model] = models.Model
    fields: List = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.instance: Optional[models.Model] = None

    def get_queryset(self) -> Optional[models.QuerySet]:
        """Gets the queryset in order to be able to access to annotated fields."""
        return self.queryset

    def get_object(self, pk: int) -> models.Model:
        """Gets the object, using the queryset if provided to add annotation fields."""
        queryset = self.get_queryset()
        if not queryset:
            return get_object_or_404(self.model_class, pk=pk)
        instance = queryset.filter(pk=pk).first()
        if not instance:
            raise Http404
        return instance

    def get(self, request, pk):
        self.instance = self.get_object(pk=pk)
        context = {"instance": self.instance, "fields": self.fields}
        context.update(self.get_extra_context())
        return render(request, self.template_name, context=context)


class BackOfficeIndexView(BackOfficeViewMixin, View):
    """Home view of the backoffice_extensions."""

    template_name = "backoffice/index.html"
    sign_in_redirect: AnyStr = f"{URL_NAMESPACE}:sign-in"

    @staticmethod
    def default_queryset() -> Dict:
        """Default queryset to each model used in statistics."""
        return {}

    def get_context_data(self) -> Dict:
        """Overwrite to add context to the view."""
        return {}

    def get(self, request):
        if not request.user.is_authenticated:
            return redirect(self.sign_in_redirect)
        context = self.get_context_data()
        context.update(self.get_extra_context())
        return render(request, self.template_name, context=context)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['backoffice_extensions', 'backoffice_extensions.templatetags']

package_data = \
{'': ['*'],
 'backoffice_extensions': ['templates/backoffice/*',
                           'templates/backoffice/bases/*',
                           'templates/backoffice/partials/*']}

install_requires = \
['django-model-utils>=4.0.0,<5.0.0',
 'django>=3.0.7,<4.0.0',
 'pytest-cov>=2.10.0,<3.0.0']

setup_kwargs = {
    'name': 'django-backoffice-extensions',
    'version': '1.0.1',
    'description': 'A set of views, tools and helpers to create a backoffice using Django.',
    'long_description': None,
    'author': 'Marcos Gabarda',
    'author_email': 'marcos@dekalabs.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Dekalabs/django-backoffice-extensions',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
